use core::fmt::Display;

use arrayvec::ArrayString;
use crc::{Crc, CRC_32_ISCSI};
use embedded_storage::nor_flash::{NorFlash, ReadNorFlash};
use serde::{Deserialize, Serialize};
use stm32f4xx_hal::flash::{FlashExt, LockedFlash};

use crate::geo::distance_km;

const DEFAULT_FREQUENCY: u32 = 430_500_000;
const CONFIG_LEN: usize = 2048;
const CASTAGNOLI: Crc<u32> = Crc::<u32>::new(&CRC_32_ISCSI);

#[derive(Clone, Serialize, Deserialize)]
pub struct Config {
    pub frequency: Frequency,
    pub callsign: ArrayString<252>,
    pub icon: u16,
    pub ssid: u8,
    pub max_hops: u8,
    pub comment: ArrayString<512>,
    pub transmit_period_seconds: u64,
    pub gps: GpsSetting,
    pub enable_transmit: bool,
    pub enable_digipeating: bool,
    pub black_hole: MaybeBlackHole,
    pub min_voltage: Maybe<f32>,

    pub antenna_height: Maybe<u8>,
    pub antenna_gain: Maybe<f32>,
    pub tx_power: Maybe<f32>,
}

impl Config {
    pub fn save_to_flash(&self, flash: &mut LockedFlash) {
        let mut unlocked = flash.unlocked();

        // Erase last flash sector (128kB)
        NorFlash::erase(&mut unlocked, 384 * 1024, 512 * 1024).unwrap();

        // Write our config
        let mut buf = [0; CONFIG_LEN];
        self.serialize(&mut buf);
        NorFlash::write(&mut unlocked, 384 * 1024, &buf).unwrap();
    }

    pub fn load_from_flash(flash: &mut LockedFlash) -> Option<Self> {
        let mut buf = [0; CONFIG_LEN];
        ReadNorFlash::read(flash, 384 * 1024, &mut buf).unwrap();

        Self::deserialize(&buf)
    }

    pub fn should_enable_gps(&self) -> bool {
        self.enable_transmit
            && (matches!(self.black_hole, MaybeBlackHole::Some(_))
                || matches!(self.gps, GpsSetting::Receiver))
    }

    fn serialize(&self, buf: &mut [u8; CONFIG_LEN]) {
        // version
        // allows us to update the config structure in the future
        // while maintaining backwards compatibility
        buf[0] = 1;

        postcard::to_slice(self, &mut buf[1..]).unwrap();

        let checksum = CASTAGNOLI.checksum(&buf[0..2044]);
        buf[2044..2048].copy_from_slice(&checksum.to_le_bytes());
    }

    fn deserialize(buf: &[u8; CONFIG_LEN]) -> Option<Self> {
        match buf[0] {
            0 => Self::deserialize_v0(buf),
            1 => Self::deserialize_v1(buf),
            _ => None,
        }
    }

    fn deserialize_v1(buf: &[u8; CONFIG_LEN]) -> Option<Self> {
        let expected_checksum = CASTAGNOLI.checksum(&buf[0..2044]);
        let actual_checksum = u32::from_le_bytes(buf[2044..2048].try_into().unwrap());
        if expected_checksum != actual_checksum {
            return None;
        }

        postcard::from_bytes(&buf[1..2044]).ok()
    }

    fn deserialize_v0(buf: &[u8; CONFIG_LEN]) -> Option<Self> {
        let expected_checksum = CASTAGNOLI.checksum(&buf[0..840]);
        let actual_checksum = u32::from_le_bytes(buf[840..844].try_into().unwrap());
        if expected_checksum != actual_checksum {
            return None;
        }

        let icon = u16::from_le_bytes(buf[1..3].try_into().unwrap());
        let ssid = buf[3];
        let max_hops = buf[4];
        let transmit_period_seconds = u64::from_le_bytes(buf[5..13].try_into().unwrap());
        let gps = match buf[13] {
            0 => GpsSetting::Disabled,
            1 => {
                let lat = f64::from_le_bytes(buf[14..22].try_into().unwrap());
                let lon = f64::from_le_bytes(buf[22..30].try_into().unwrap());

                GpsSetting::Fixed(lat, lon)
            }
            2 => GpsSetting::Receiver,
            _ => return None,
        };

        let enable_transmit = (buf[30] & 1) > 0;
        let enable_digipeating = (buf[30] & 2) > 0;

        let callsign_len: usize = buf[31].into();
        let callsign = core::str::from_utf8(&buf[32..(32 + callsign_len)]).ok()?;
        let callsign = ArrayString::from(callsign).ok()?;

        let comment_len: usize = u16::from_le_bytes(buf[284..286].try_into().unwrap()).into();
        if comment_len > 512 {
            return None;
        }
        let comment = core::str::from_utf8(&buf[286..(286 + comment_len)]).ok()?;
        let comment = ArrayString::from(comment).ok()?;

        let black_hole = MaybeBlackHole::decode(&buf[798..823].try_into().unwrap())?;

        let antenna_height = if buf[823] > 0 {
            Maybe::Some(buf[824])
        } else {
            Maybe::None
        };

        let antenna_gain = if buf[825] > 0 {
            Maybe::Some(f32::from_le_bytes(buf[826..830].try_into().unwrap()))
        } else {
            Maybe::None
        };

        let tx_power = if buf[830] > 0 {
            Maybe::Some(f32::from_le_bytes(buf[831..835].try_into().unwrap()))
        } else {
            Maybe::None
        };

        let min_voltage = if buf[835] > 0 {
            Maybe::Some(f32::from_le_bytes(buf[836..840].try_into().unwrap()))
        } else {
            Maybe::None
        };

        Some(Config {
            frequency: Frequency::default(),
            callsign,
            icon,
            ssid,
            max_hops,
            comment,
            transmit_period_seconds,
            gps,
            enable_transmit,
            enable_digipeating,
            black_hole,
            min_voltage,

            antenna_height,
            antenna_gain,
            tx_power,
        })
    }
}

impl Default for Config {
    fn default() -> Self {
        Config {
            frequency: Frequency::default(),
            callsign: "N0CALL".try_into().unwrap(),
            icon: 0,
            ssid: 0,
            max_hops: 2,
            comment: "".try_into().unwrap(),
            transmit_period_seconds: 60,
            gps: GpsSetting::Disabled,
            enable_transmit: false,
            enable_digipeating: true,
            black_hole: MaybeBlackHole::None,
            min_voltage: Maybe::None,

            antenna_height: Maybe::None,
            antenna_gain: Maybe::None,
            tx_power: Maybe::Some(30.0),
        }
    }
}

#[derive(Copy, Clone, Serialize, Deserialize)]
pub enum GpsSetting {
    Disabled,
    Fixed(f64, f64),
    Receiver,
}

impl Display for GpsSetting {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        match self {
            GpsSetting::Disabled => write!(f, "disabled"),
            GpsSetting::Fixed(lat, lon) => write!(f, "fixed {lat:.6} {lon:.6}"),
            GpsSetting::Receiver => write!(f, "receiver"),
        }
    }
}

#[derive(Copy, Clone, Serialize, Deserialize)]
pub enum MaybeBlackHole {
    Some(BlackHoleSetting),
    None,
}

impl MaybeBlackHole {
    pub fn in_black_hole(&self, lat: f64, lon: f64) -> bool {
        match self {
            MaybeBlackHole::Some(bhs) => {
                let dist_km = distance_km(lat, lon, bhs.latitude, bhs.longitude);

                dist_km * 1000.0 < bhs.radius_meters
            }
            MaybeBlackHole::None => false,
        }
    }

    fn decode(buf: &[u8; 25]) -> Option<Self> {
        match buf[0] {
            0 => Some(Self::None),
            1 => Some(MaybeBlackHole::Some(BlackHoleSetting::decode(
                buf[1..].try_into().unwrap(),
            ))),
            _ => None,
        }
    }
}

impl Display for MaybeBlackHole {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        match self {
            MaybeBlackHole::Some(bhs) => write!(f, "{bhs}"),
            MaybeBlackHole::None => write!(f, "<None>"),
        }
    }
}

#[derive(Copy, Clone, Serialize, Deserialize)]
pub struct BlackHoleSetting {
    pub latitude: f64,
    pub longitude: f64,
    pub radius_meters: f64,
}

impl BlackHoleSetting {
    fn decode(buf: &[u8; 24]) -> Self {
        let latitude = f64::from_le_bytes(buf[0..8].try_into().unwrap());
        let longitude = f64::from_le_bytes(buf[8..16].try_into().unwrap());
        let radius_meters = f64::from_le_bytes(buf[16..24].try_into().unwrap());

        Self {
            latitude,
            longitude,
            radius_meters,
        }
    }
}

impl Display for BlackHoleSetting {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        write!(
            f,
            "{:.6} {:.6} {:.2}",
            self.latitude, self.longitude, self.radius_meters
        )
    }
}

// Like Option, but local, so we can implement Display on it
#[derive(Copy, Clone, Serialize, Deserialize)]
pub enum Maybe<T> {
    Some(T),
    None,
}

impl<T> Display for Maybe<T>
where
    T: Display,
{
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        match self {
            Maybe::Some(x) => x.fmt(f),
            Maybe::None => write!(f, "<None>"),
        }
    }
}

impl<T> From<Maybe<T>> for Option<T> {
    fn from(val: Maybe<T>) -> Self {
        match val {
            Maybe::Some(x) => Some(x),
            Maybe::None => None,
        }
    }
}

#[derive(Copy, Clone, Serialize, Deserialize)]
pub struct Frequency(u32);

impl Frequency {
    pub fn new(f: u32) -> Option<Self> {
        if Self::validate_freq(f) {
            Some(Self(f))
        } else {
            None
        }
    }

    pub fn new_mhz(f: f64) -> Option<Self> {
        let hz = u32::try_from((f * 1_000_000.0) as u64).ok()?;

        Self::new(hz)
    }

    pub fn hz(&self) -> u32 {
        self.0
    }

    fn validate_freq(f: u32) -> bool {
        (420_000_000..=450_000_000).contains(&f)
    }
}

impl Default for Frequency {
    fn default() -> Self {
        Self(DEFAULT_FREQUENCY)
    }
}

impl Display for Frequency {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        let freq_mhz = self.0 as f64 / 1_000_000.0;

        write!(f, "{freq_mhz:.3} MHz")
    }
}
